#MIT License
#
#Copyright (c) 2021 Phil Wang

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
# =========================================================================

import torch.nn as nn

class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(x, **kwargs) + x
    
class TargetLengthCrop(nn.Module):
    def __init__(self, target_length):
        super().__init__()
        self.target_length = target_length

    def forward(self, x):
        seq_len, target_len = x.shape[-2], self.target_length

        if target_len == -1:
            return x

        if seq_len < target_len:
            raise ValueError(f'sequence length {seq_len} is less than target length {target_len}')

        trim = (target_len - seq_len) // 2

        if trim == 0:
            return x

        return x[:, -trim:trim]


def undo_squashed_scale(x, clip_soft=384, track_transform=3 / 4, track_scale = 0.01, old_transform = True):
    """
    Reverses the squashed scaling transformation applied to the output profiles.

    Args:
        x (torch.Tensor): The input tensor to be unsquashed.
        clip_soft (float, optional): The soft clipping value. Defaults to 384.
        track_transform (float, optional): The transformation factor. Defaults to 3/4.
        track_scale (float, optional): The scale factor. Defaults to 0.01.

    Returns:
        torch.Tensor: The unsquashed tensor.
    """
    x = x.clone()  # IMPORTANT BECAUSE OF IMPLACE OPERATIONS TO FOLLOW?
    
    if old_transform:
        x = x / track_scale
        unclip_mask = x > clip_soft
        x[unclip_mask] = (x[unclip_mask] - clip_soft) ** 2 + clip_soft
        x = x ** (1./track_transform)
    else:
        unclip_mask = x > clip_soft
        x[unclip_mask] = (x[unclip_mask] - clip_soft+1) ** 2 + clip_soft -1
        x = (x + 1) ** (1.0 / track_transform) - 1
        x = x / track_scale
    return x


#Adapted from https://github.com/lucidrains/enformer-pytorch/tree/main
#
#MIT License
#
#Copyright (c) 2021 Phil Wang, 2024 Johannes Hingerl

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
# =========================================================================

import torch.nn as nn
from torch import nn, einsum
from einops import rearrange, reduce
from einops.layers.torch import Rearrange
import torch
import math


def get_positional_features_central_mask(positions, features, seq_len):
    pow_rate = math.exp(math.log(seq_len + 1) / features)
    center_widths = torch.pow(pow_rate, torch.arange(1, features + 1, device = positions.device)).float()
    center_widths = center_widths - 1
    return (center_widths[None, ...] > positions.abs()[..., None]).float()


def get_positional_embed(seq_len, feature_size, device):
    distances = torch.arange(-seq_len + 1, seq_len, device = device)

    feature_functions = [
        get_positional_features_central_mask,
    ]

    num_components = len(feature_functions) * 2

    if (feature_size % num_components) != 0:
        raise ValueError(f'feature size is not divisible by number of components ({num_components})')

    num_basis_per_class = feature_size // num_components

    embeddings = []
    for fn in feature_functions:
        embeddings.append(fn(distances, num_basis_per_class, seq_len))

    embeddings = torch.cat(embeddings, dim = -1)
    embeddings = torch.cat((embeddings, torch.sign(distances)[..., None] * embeddings), dim = -1)
    return embeddings

def fast_relative_shift(a,b):
    return einsum("i d, j d -> i j", a, b).flatten().as_strided(size =(a.shape[0],a.shape[0]), stride= ((a.shape[0]-1)*2,1), storage_offset = a.shape[0] - 1)

fast_relative_shift= torch.vmap(torch.vmap(fast_relative_shift), in_dims=(0, None)) #https://johahi.github.io/blog/2024/fast-relative-shift/


class Attention(nn.Module):
    
    def __init__(
        self,
        dim=1536,
        *,
        num_rel_pos_features = 1,
        heads = 8,
        dim_key = 64,
        dim_value = 64,
        dropout = 0.,
        pos_dropout = 0.
    ):
        super().__init__()
        self.scale = dim_key ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(dim, dim_key * heads, bias = False)
        self.to_k = nn.Linear(dim, dim_key * heads, bias = False)
        self.to_v = nn.Linear(dim, dim_value * heads, bias = False)

        self.to_out = nn.Linear(dim_value * heads, dim)
        nn.init.zeros_(self.to_out.weight)
        nn.init.zeros_(self.to_out.bias)

        # relative positional encoding

        self.num_rel_pos_features = num_rel_pos_features
        
        self.register_buffer("positions",get_positional_embed(4096, self.num_rel_pos_features, self.to_v.weight.device), persistent = False) # 4096 as this should always be the seq len at this pos?

        self.to_rel_k = nn.Linear(num_rel_pos_features, dim_key * heads, bias = False)
        self.rel_content_bias = nn.Parameter(torch.randn(1, heads, 1, dim_key))
        self.rel_pos_bias = nn.Parameter(torch.randn(1, heads, 1, dim_key))

        # dropouts

        self.pos_dropout = nn.Dropout(pos_dropout)
        self.attn_dropout = nn.Dropout(dropout)

    def forward(self, x):
        n, h, device = x.shape[-2], self.heads, x.device
        
        q = self.to_q(x)
        k = self.to_k(x)
        v = self.to_v(x)

        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))

        q = q * self.scale

        content_logits = einsum('b h i d, b h j d -> b h i j', q + self.rel_content_bias, k)

        positions = self.pos_dropout(self.positions)
        rel_k = self.to_rel_k(positions)
        rel_k = rearrange(rel_k, 'n (h d) -> h n d', h = h)
        rel_logits = fast_relative_shift(q + self.rel_pos_bias,rel_k)
        logits = content_logits + rel_logits
        attn = logits.softmax(dim = -1)
        attn = self.attn_dropout(attn)

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h n d -> b n (h d)')
        out = self.to_out(out)
        return out
    
class FlashAttention(nn.Module):
    def __init__(
        self,
        dim=1536,
        heads = 8,
        dropout = 0.15,
        pos_dropout = 0.15, # Not used
        rotary_emb_base = 20000.0,
        rotary_emb_scale_base = None, 
        ):
        super().__init__()

        from flash_attn.modules.mha import MHA
        self.mha = MHA(
            use_flash_attn=True,
            embed_dim=dim,
            num_heads = heads,
            num_heads_kv = (heads//2),
            qkv_proj_bias=True,#False,
            out_proj_bias=True,
            dropout=dropout,
            softmax_scale=(dim/heads) ** -0.5,
            causal=False,
            rotary_emb_dim=128,
            rotary_emb_base=rotary_emb_base,
            rotary_emb_scale_base = rotary_emb_scale_base,
            fused_bias_fc = False,
        ) 

        nn.init.kaiming_normal_(self.mha.Wqkv.weight, nonlinearity = 'relu')
        nn.init.zeros_(self.mha.out_proj.weight)
        nn.init.zeros_(self.mha.out_proj.bias)
        nn.init.ones_(self.mha.Wqkv.bias)


    def forward(self, x):
        out = self.mha(x)
        return out



from transformers import PretrainedConfig

class BorzoiConfig(PretrainedConfig):
    model_type = "borzoi"

    def __init__(
        self,
        dim = 1536,
        depth = 8,
        heads = 8,
        # output_heads = dict(human = 5313, mouse= 1643),
        return_center_bins_only = True,
        attn_dim_key = 64,
        attn_dim_value = 192,
        dropout_rate = 0.2,
        attn_dropout = 0.05,
        pos_dropout = 0.01,
        enable_mouse_head = False,
        **kwargs,
    ):
        self.dim = dim
        self.depth = depth
        self.heads = heads
        # self.output_heads = output_heads
        self.attn_dim_key = attn_dim_key
        self.attn_dim_value = attn_dim_value
        self.dropout_rate = dropout_rate
        self.attn_dropout = attn_dropout
        self.pos_dropout = pos_dropout
        self.return_center_bins_only = return_center_bins_only
        self.enable_mouse_head = enable_mouse_head
        super().__init__(**kwargs)


# Copyright 2023 Calico LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========================================================================



from borzoi_pytorch.config_borzoi import BorzoiConfig
from transformers import PreTrainedModel
import torch.nn as nn
import torch
import numpy as np
import math
import copy
from pathlib import Path

from .pytorch_borzoi_utils import Residual, TargetLengthCrop, undo_squashed_scale
from .pytorch_borzoi_transformer import Attention, FlashAttention

import pandas as pd
DIR = Path(__file__).parents[0]
TRACKS_DF = pd.read_table(str(DIR / "precomputed"/ "targets.txt")).rename(columns={'Unnamed: 0':'index'})

#torch.backends.cudnn.deterministic = True

#torch.set_float32_matmul_precision('high')
  
class ConvDna(nn.Module):
    def __init__(self):
        super(ConvDna, self).__init__()
        self.conv_layer = nn.Conv1d(in_channels = 4,out_channels = 512, kernel_size = 15, padding="same")
        self.max_pool = nn.MaxPool1d(kernel_size = 2, padding = 0)

    def forward(self, x):
        return self.max_pool(self.conv_layer(x))

class ConvBlock(nn.Module):
    def __init__(self, in_channels,out_channels=None, kernel_size=1,
                 conv_type="standard"):
        super(ConvBlock, self).__init__()
        if conv_type == "separable":
            self.norm = nn.Identity()
            depthwise_conv = nn.Conv1d(in_channels, out_channels, kernel_size=kernel_size, groups=in_channels, padding = 'same', bias = False)
            pointwise_conv = nn.Conv1d(in_channels, out_channels, kernel_size=1)
            self.conv_layer = nn.Sequential(depthwise_conv, pointwise_conv)
            self.activation = nn.Identity()
        else:
            self.norm = nn.BatchNorm1d(in_channels, eps = 0.001)
            self.activation =nn.GELU(approximate='tanh')
            self.conv_layer = nn.Conv1d(
                in_channels, 
                out_channels, 
                kernel_size=kernel_size,
                padding='same')    
            
    def forward(self, x):
        x = self.norm(x)
        x = self.activation(x)
        x = self.conv_layer(x)
        return x

    
class Borzoi(PreTrainedModel):
    config_class = BorzoiConfig
    base_model_prefix = "borzoi"

    @staticmethod
    def from_hparams(**kwargs):
        return Borzoi(BorzoiConfig(**kwargs))
    
    
    def __init__(self, config):
        super(Borzoi, self).__init__(config)
        self.flashed = config.flashed if "flashed" in config.__dict__.keys() else False
        self.enable_human_head = config.enable_human_head if "enable_human_head" in config.__dict__.keys() else True
        self.enable_mouse_head = config.enable_mouse_head   
        self.conv_dna = ConvDna()
        self._max_pool = nn.MaxPool1d(kernel_size = 2, padding = 0)
        self.res_tower = nn.Sequential(
            ConvBlock(in_channels = 512,out_channels = 608,kernel_size = 5),
            self._max_pool,
            ConvBlock(in_channels = 608,out_channels = 736,kernel_size = 5),
            self._max_pool,
            ConvBlock(in_channels = 736,out_channels = 896,kernel_size = 5),
            self._max_pool,
            ConvBlock(in_channels = 896,out_channels = 1056,kernel_size = 5),
            self._max_pool,
            ConvBlock(in_channels = 1056,out_channels = 1280,kernel_size = 5),
        )
        self.unet1 = nn.Sequential(
            self._max_pool,
            ConvBlock(in_channels = 1280,out_channels = config.dim,kernel_size = 5),
        )
        transformer = []
        for _ in range(config.depth):
            transformer.append(nn.Sequential(
                Residual(nn.Sequential(
                    nn.LayerNorm(config.dim, eps = 0.001),
                    Attention(
                        config.dim,
                        heads = config.heads,
                        dim_key = config.attn_dim_key,
                        dim_value = config.attn_dim_value,
                        dropout = config.attn_dropout,
                        pos_dropout = config.pos_dropout,
                        num_rel_pos_features = 32
                    ) if not self.flashed else
                    FlashAttention(
                        config.dim,
                        heads = config.heads,
                        dropout = config.attn_dropout,
                        pos_dropout = config.pos_dropout,
                    ),
                    nn.Dropout(0.2))
                ),
                Residual(nn.Sequential(
                    nn.LayerNorm(config.dim, eps = 0.001),
                    nn.Linear(config.dim, config.dim * 2),
                    nn.Dropout(config.dropout_rate),
                    nn.ReLU(),
                    nn.Linear(config.dim * 2, config.dim),
                    nn.Dropout(config.dropout_rate)
                )))
            )
        self.horizontal_conv0,self.horizontal_conv1 = ConvBlock(in_channels = 1280, out_channels = config.dim, kernel_size = 1),ConvBlock(in_channels = config.dim, out_channels = config.dim,kernel_size = 1)
        self.upsample = torch.nn.Upsample(scale_factor = 2)
        self.transformer = nn.Sequential(*transformer)
        self.upsampling_unet1 = nn.Sequential(
            ConvBlock(in_channels = config.dim, out_channels = config.dim,  kernel_size = 1),
            self.upsample,
        )
        self.separable1 = ConvBlock(in_channels = config.dim, out_channels = config.dim,  kernel_size = 3, conv_type = 'separable')
        self.upsampling_unet0 = nn.Sequential(
            ConvBlock(in_channels = config.dim,out_channels = config.dim,kernel_size = 1),
            self.upsample,
        )
        self.separable0 = ConvBlock(in_channels = config.dim, out_channels = config.dim,  kernel_size = 3, conv_type = 'separable')
        if config.return_center_bins_only:
            self.crop = TargetLengthCrop(6144)
        else:
            self.crop = TargetLengthCrop(16384 - 32) # as in Borzoi     
        self.final_joined_convs = nn.Sequential(
            ConvBlock(in_channels = config.dim, out_channels = 1920, kernel_size = 1),
            nn.Dropout(0.1),
            nn.GELU(approximate='tanh'),
        )
        if self.enable_human_head:
            self.human_head = nn.Conv1d(in_channels = 1920, out_channels = 7611, kernel_size = 1)
        if self.enable_mouse_head:
            self.mouse_head = nn.Conv1d(in_channels = 1920, out_channels = 2608, kernel_size = 1)
        self.final_softplus = nn.Softplus()


    def _init_weights(self, module):
        """ Initialize the weights """
        if isinstance(module, (nn.Linear, nn.Embedding, nn.Conv1d)):
            # module.weight.data.normal_(mean=0.0, std=0.02)
            nn.init.xavier_normal_(module.weight)
        elif isinstance(module, (nn.LayerNorm, nn.BatchNorm1d)):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, (nn.Linear, nn.Conv1d)) and module.bias is not None:
            module.bias.data.zero_()

    
    def set_track_subset(self, track_subset):
        if not hasattr(self, 'human_head_bak'):
            self.human_head_bak = copy.deepcopy(self.human_head)
        else:
            self.reset_track_subset()
        self.human_head = nn.Conv1d(1920, len(track_subset), 1)
        self.human_head.weight = nn.Parameter(self.human_head_bak.weight[track_subset].clone())
        self.human_head.bias = nn.Parameter(self.human_head_bak.bias[track_subset].clone())

    
    def reset_track_subset(self):
        self.human_head = copy.deepcopy(self.human_head_bak)

    
    def get_embs_after_crop(self, x):
        x = self.conv_dna(x)
        x_unet0 = self.res_tower(x)
        x_unet1 = self.unet1(x_unet0)
        x = self._max_pool(x_unet1)
        x_unet1 = self.horizontal_conv1(x_unet1)
        x_unet0 = self.horizontal_conv0(x_unet0)
        x = self.transformer(x.permute(0,2,1))
        x = x.permute(0,2,1)
        x = self.upsampling_unet1(x)
        x += x_unet1
        x = self.separable1(x)
        x = self.upsampling_unet0(x)
        x += x_unet0
        x = self.separable0(x)
        x = self.crop(x.permute(0,2,1))
        return x.permute(0,2,1)

    
    def predict(self, seqs, gene_slices, remove_squashed_scale = False):
        """
        Predicts only for bins of interest in a batched fashion
        Args:
            seqs (torch.tensor): Nx4xL tensor of one-hot sequences
            gene_slices List[torch.Tensor]: tensors indicating bins of interest
            removed_squashed_scale (bool, optional): whether to undo the squashed scale

        Returns:
            Tuple[torch.Tensor, list[int]]: 1xCxB tensor of bin predictions, as well as offsets that indicate where sequences begin/end
        """
        # Calculate slice offsets
        slice_list = []
        slice_length = []
        offset = self.crop.target_length
        for i,gene_slice in enumerate(gene_slices):
            slice_list.append(gene_slice + i*offset)
            slice_length.append(gene_slice.shape[0])
        slice_list = torch.concatenate(slice_list)
        # Get embedding after cropped 
        seq_embs = self.get_embs_after_crop(seqs)
        # Reshape to flatten the batch dimension (i.e. concatenate sequences)
        seq_embs = seq_embs.permute(1,0,2).flatten(start_dim=1).unsqueeze(0)
        # Extract the bins of interest
        seq_embs = seq_embs[:,:,slice_list]
        # Run the model head
        seq_embs = self.final_joined_convs(seq_embs)
        with torch.cuda.amp.autocast(enabled = False):
            conved_slices = self.final_softplus(self.human_head(seq_embs.float()))
        if remove_squashed_scale:
            conved_slices = undo_squashed_scale(conved_slices)
        return conved_slices, slice_length


    def forward(self, x, is_human = True, data_parallel_training = False):
        x = self.get_embs_after_crop(x)
        x = self.final_joined_convs(x)
        # disable autocast for more precision in final layer
        with torch.cuda.amp.autocast(enabled=False):
            if data_parallel_training:
                # we need this to get gradients for both heads if doing DDP training
                if is_human:
                    human_out = self.final_softplus(self.human_head(x.float())) + 0 * self.mouse_head(x.float()).sum()
                    return human_out
                else:
                    mouse_out = self.final_softplus(self.mouse_head(x.float())) + 0 * self.human_head(x.float()).sum()
                    return mouse_out
            else:
                if is_human:
                    return self.final_softplus(self.human_head(x.float()))
                else:
                    return self.final_softplus(self.mouse_head(x.float()))


class AnnotatedBorzoi(Borzoi):
    
    def __init__(self, config, tracks_df=TRACKS_DF):
        super(AnnotatedBorzoi, self).__init__(config)
        assert all(x in tracks_df.columns for x in ['identifier', 'file', 'clip', 'clip_soft', 'scale', 'sum_stat', 'strand_pair', 'description'])
        tracks_df['track_transform'] = tracks_df['sum_stat'].apply(lambda x: 3/4 if x == "sum_sqrt" else 1.)
        self._build_annotation_df(tracks_df)

    def _build_annotation_df(self,tracks_df):
        # build tensor of tracks (sense, antisense, unstranded)
        self.sense_tracks = torch.tensor(tracks_df.loc[tracks_df.identifier.str.contains('\+') | (tracks_df.index == tracks_df['strand_pair'])].index)
        self.antisense_tracks = torch.tensor(tracks_df.loc[tracks_df.identifier.str.endswith('-') | (tracks_df.index == tracks_df['strand_pair'])].index)
        # check that ordering of sense and antisense is meaningful
        assert ((tracks_df.iloc[self.antisense_tracks].description.array == tracks_df.iloc[self.sense_tracks].description.array).sum() == self.sense_tracks.shape[0])
        # remember backing dataframe
        self.tracks_df = tracks_df
        self.output_tracks_df = tracks_df.loc[tracks_df.identifier.str.contains('\+') | (tracks_df.index == tracks_df['strand_pair'])].reset_index(drop=True)
        self.register_buffer('scale_values', torch.from_numpy(self.output_tracks_df.scale.values).float().unsqueeze(0).unsqueeze(-1).to(self.conv_dna.conv_layer.weight.device), persistent=False)
        self.register_buffer('clip_values', torch.from_numpy(self.output_tracks_df.clip_soft.values).float().unsqueeze(0).unsqueeze(-1).to(self.conv_dna.conv_layer.weight.device), persistent=False)
        self.register_buffer('track_transform', torch.from_numpy(self.output_tracks_df.track_transform.values).float().unsqueeze(0).unsqueeze(-1).to(self.conv_dna.conv_layer.weight.device), persistent=False)

    def set_track_subset(self, track_subset):
        if not hasattr(self, 'tracks_df_bak'):
            tracks_df = self.tracks_df.copy()
            self.tracks_df_bak = tracks_df
        else:
            tracks_df = self.tracks_df_bak.copy()
        tracks_df = tracks_df.iloc[track_subset].reset_index(names='old_index')
        # remap indices and strand pairs
        oldidx_to_newidx = {x['old_index']:i for i,x in tracks_df.iterrows()}
        try:
            tracks_df["strand_pair"] = tracks_df['strand_pair'].apply(lambda x: oldidx_to_newidx[x])
        except KeyError as e: 
            raise Exception("Strand pair is missing")
        # subset head
        super().set_track_subset(track_subset)
        # rebuild annotation
        tracks_df = tracks_df.drop(columns='old_index')
        self._build_annotation_df(tracks_df)

    def reset_track_subset(self):
        tracks_df = self.tracks_df_bak.copy()
        super.reset_track_subset()
        self._build_annotation_df(tracks_df)
    
    def _predict_gene_count(self, x, 
                gene_slices, 
                predict_antisense = False
               ):
        if predict_antisense:
            # revcomp the input
            x = x.flip(dims=(1,2))
            # adapt the bins
            gene_slices = [(self.crop.target_length - 1 - slices) for slices in gene_slices]
        # run model
        x, slice_length = self.predict(x, gene_slices = gene_slices)
        if predict_antisense:
            # extract the antisense and unstranded tracks
            x = x[:,self.antisense_tracks,:]
        else:
            # extract the sense and unstranded tracks
            x = x[:,self.sense_tracks,:]
        return x, slice_length

    def _undo_squashed_scale(self,x, old_transform = True, unscale = True):
        """
        Reverses the squashed scaling transformation applied to the output profiles.
        Uses the annotation df to supply information how this should be done.
    
        Args:
            x (torch.Tensor): The input tensor to be unsquashed.
            old_transform: Which version of the transform to use
        Returns:
            torch.Tensor: The unsquashed tensor.
        """
        x = x.clone()  # IMPORTANT BECAUSE OF IMPLACE OPERATIONS TO FOLLOW?

        clip_soft = self.clip_values.expand_as(x)
        track_transform = self.track_transform.expand_as(x)
        if unscale:
            scale = self.scale_values.expand_as(x)
        else:
            scale = 1.
        
        if old_transform:
            x = x / scale
            unclip_mask = x > clip_soft
            x[unclip_mask] = (x[unclip_mask] - clip_soft[unclip_mask]) ** 2 + clip_soft[unclip_mask]
            x = x ** (1./track_transform)
        else:
            unclip_mask = x > clip_soft
            x[unclip_mask] = (x[unclip_mask] - clip_soft[unclip_mask] + 1) ** 2 + clip_soft[unclip_mask] -1
            x = (x + 1) ** (1.0 / track_transform) - 1
            x = x / scale
        return x

    def predict_gene_count(self, x, 
                gene_slices = None, 
                average_strands = True,
                bin_level_transform = None,
                agg_fn = lambda x: torch.sum(x, dim=-1),
                log1p = True
               ):
        if gene_slices is None: # predict everything if no slices provided
            gene_slices = [torch.tensor([x for x in range(self.crop.target_length)]) for i in range(x.shape[0])]
        pred_sense, slice_length = self._predict_gene_count(x, gene_slices)
        if average_strands:
            pred_antisense, slice_length = self._predict_gene_count(x, gene_slices, predict_antisense = True)
            pred = (pred_sense + pred_antisense)/2
        else:
            pred = pred_sense
        # sum, unsquash and log1p-transform
        if bin_level_transform is None:
            pred = self._undo_squashed_scale(pred)
        else:
            pred = bin_level_transform(pred)
        pred = torch.stack([agg_fn(x[0]) for x in torch.split(pred, slice_length, dim = 2)])
        if log1p:
            pred = torch.log1p(pred)
        return pred



import numpy as np
import torch

def predict_tracks(models, sequence_one_hot, slices):
    predicted_tracks = []
    for fold_ix in range(len(models)):
        with torch.no_grad():
            yh = models[fold_ix](sequence_one_hot[None, ...])[:, None, ...].numpy(force = True)[:,:,slices]
        predicted_tracks.append(yh)

    predicted_tracks = np.concatenate(predicted_tracks,axis=1).swapaxes(3,2)

    return predicted_tracks


from setuptools import setup, find_packages

setup(
    name='borzoi-pytorch',
    version='0.4.0',
    author='Johannes Hingerl',
    author_email='johannes.hingerl@tum.de',
    packages=['borzoi_pytorch'],
    # packages = find_packages(exclude=[]),
    include_package_data = True,
    url='https://github.com/johahi/borzoi-pytorch',
    license='LICENSE',
    description='The Borzoi model from Linder et al., but in Pytorch',
    install_requires=[
        "einops >= 0.5",
        "numpy >= 1.14.2",
        "torch >= 2.1.0",
        "transformers >= 4.34.1",
        "jupyter >= 1.0.0",
	"intervaltree~=3.1.0",
	"pandas",
        "flash-attn >= 2.6.3; extra == 'flash'"
    ],
)



# Copyright 2022 Calico LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========================================================================

# adapted from calico/baskerville

import gzip
from intervaltree import IntervalTree
import numpy as np


class GenomicInterval:
    def __init__(self, start, end, chrom=None, strand=None):
        self.start = start
        self.end = end
        self.chrom = chrom
        self.strand = strand

    def __eq__(self, other):
        return self.start == other.start

    def __lt__(self, other):
        return self.start < other.start

    def __cmp__(self, x):
        if self.start < x.start:
            return -1
        elif self.start > x.start:
            return 1
        else:
            return 0

    def __str__(self):
        if self.chrom is None:
            label = "[%d-%d]" % (self.start, self.end)
        else:
            label = "%s:%d-%d" % (self.chrom, self.start, self.end)
        return label


class Gene:
    """Class for managing genes in an isoform-agnostic way, taking
    the union of exons across isoforms."""

    def __init__(self, chrom, strand, kv, name=None):
        self.chrom = chrom
        self.strand = strand
        self.kv = kv
        self.name = name
        self.exons = IntervalTree()

    def add_exon(self, start, end):
        """BED 0-indexing assumed."""
        self.exons[start:end] = True

    def get_exons(self):
        self.exons.merge_overlaps()
        return sorted(self.exons)

    def midpoint(self):
        positions = []
        for exon in self.get_exons():
            positions += range(exon.begin, exon.end)
        midp = int(np.mean(positions))
        return midp

    def span(self):
        exon_starts = [exon.begin for exon in self.exons]
        exon_ends = [exon.end for exon in self.exons]
        return min(exon_starts), max(exon_ends)
    
    def output_slice_old(self, seq_start, seq_len, model_stride, span=False):
        gene_slice = []

        if span:
            gene_start, gene_end = self.span()

            # clip left boundaries
            gene_seq_start = max(0, gene_start - seq_start)
            gene_seq_end = max(0, gene_end - seq_start)

            # requires >50% overlap
            slice_start = int(np.round(gene_seq_start / model_stride))
            slice_end = int(np.round(gene_seq_end / model_stride))

            # clip right boundaries
            slice_max = int(seq_len/model_stride)
            slice_start = min(slice_start, slice_max)
            slice_end = min(slice_end, slice_max)

            gene_slice = range(slice_start, slice_end)

        else:
            for exon in self.get_exons():
                # clip left boundaries
                exon_seq_start = max(0, exon.begin - seq_start)
                exon_seq_end = max(0, exon.end - seq_start)

                # requires >50% overlap
                slice_start = int(np.round(exon_seq_start / model_stride))
                slice_end = int(np.round(exon_seq_end / model_stride))

                # clip right boundaries
                slice_max = int(seq_len/model_stride)
                slice_start = min(slice_start, slice_max)
                slice_end = min(slice_end, slice_max)

                gene_slice.extend(range(slice_start, slice_end))

        return np.array(gene_slice)

    def output_slice(
        self, seq_start, seq_len, model_stride, span=False, majority_overlap=False, old_version=False
    ):
        if old_version :
            return self.output_slice_old(seq_start, seq_len, model_stride, span=span)
        
        gene_slice = []

        def clip_boundaries(slice_start, slice_end):
            slice_max = int(seq_len / model_stride)
            slice_start = min(slice_start, slice_max)
            slice_end = min(slice_end, slice_max)
            slice_start = max(slice_start, 0)
            slice_end = max(slice_end, 0)
            return slice_start, slice_end

        if span:
            gene_start, gene_end = self.span()

            # clip left boundaries
            gene_seq_start = max(0, gene_start - seq_start)
            gene_seq_end = max(0, gene_end - seq_start)

            # requires >50% overlap
            slice_start = int(np.round(gene_seq_start / model_stride))
            slice_end = int(np.round(gene_seq_end / model_stride))

            # clip boundaries
            slice_start, slice_end = clip_boundaries(slice_start, slice_end)

            # add to gene slice
            if slice_start < slice_end:
                gene_slice = range(slice_start, slice_end)

        else:
            for exon in self.get_exons():
                # clip left boundaries
                exon_seq_start = max(0, exon.begin - seq_start)
                exon_seq_end = max(0, exon.end - seq_start)

                if majority_overlap:
                    # requires >50% overlap
                    slice_start = int(np.round(exon_seq_start / model_stride))
                    slice_end = int(np.round(exon_seq_end / model_stride))
                else:
                    # any overlap
                    slice_start = int(np.floor(exon_seq_start / model_stride))
                    slice_end = int(np.ceil(exon_seq_end / model_stride))

                # clip boundaries
                slice_start, slice_end = clip_boundaries(slice_start, slice_end)

                # add to gene slice
                if slice_start < slice_end:
                    gene_slice.extend(range(slice_start, slice_end))

        # collapse overlaps
        gene_slice = np.unique(gene_slice)

        return gene_slice


class Transcriptome:
    def __init__(self, gtf_file, use_geneid=False):
        self.genes = {}
        self.use_geneid = use_geneid
        self.read_gtf(gtf_file)

    def read_gtf(self, gtf_file):
        if gtf_file[-3:] == ".gz":
            gtf_in = gzip.open(gtf_file, "rt")
        else:
            gtf_in = open(gtf_file)

        # ignore header
        line = gtf_in.readline()
        while line[0] == "#":
            line = gtf_in.readline()

        while line:
            a = line.split("\t")
            if a[2] == "exon":
                chrom = a[0]
                start = int(a[3])
                end = int(a[4])
                strand = a[6]
                kv = gtf_kv(a[8])
                gene_id = kv["gene_name"] if not self.use_geneid else kv["gene_id"]

                # initialize gene
                if gene_id not in self.genes:
                    self.genes[gene_id] = Gene(chrom, strand, kv)

                # add exon
                self.genes[gene_id].add_exon(start - 1, end)

            line = gtf_in.readline()

        gtf_in.close()


################################################################################
# Methods
################################################################################
def gtf_kv(s):
    """Convert the last gtf section of key/value pairs into a dict."""
    d = {}

    a = s.split(";")
    for key_val in a:
        if key_val.strip():
            eq_i = key_val.find("=")
            if eq_i != -1 and key_val[eq_i - 1] != '"':
                kvs = key_val.split("=")
            else:
                kvs = key_val.split()

            key = kvs[0]
            if kvs[1][0] == '"' and kvs[-1][-1] == '"':
                val = (" ".join(kvs[1:]))[1:-1].strip()
            else:
                val = (" ".join(kvs[1:])).strip()

            d[key] = val

    return d
